@@ -19,6 +19,7 @@ import os
 
 import mock
 from oslo_config import cfg
+from oslo_context import context
 from oslo_serialization import jsonutils
 from oslotest import base as test_base
 import six
@@ -129,6 +130,11 @@ class RulesTestCase(test_base.BaseTestCase):
         self.assertRaises(ValueError, policy.Rules.load, exemplar,
                           'default')
 
+    @mock.patch.object(_parser, 'parse_rule', lambda x: x)
+    def test_load_empty_data(self):
+        result = policy.Rules.load('', 'default')
+        self.assertEqual(result, {})
+
     @mock.patch.object(_parser, 'parse_rule', lambda x: x)
     def test_load_yaml(self):
         # Test that simplified YAML can be used with load().
@@ -385,6 +391,66 @@ class EnforcerTest(base.PolicyBaseTestCase):
                                group='oslo_policy')
         self.assertRaises(ValueError, self.enforcer.load_rules, True)
 
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_rules_twice(self, mock_check_rules):
+        self.enforcer.load_rules()
+        self.enforcer.load_rules()
+        self.assertEqual(1, mock_check_rules.call_count)
+
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_rules_twice_force(self, mock_check_rules):
+        self.enforcer.load_rules(True)
+        self.enforcer.load_rules(True)
+        self.assertEqual(2, mock_check_rules.call_count)
+
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_rules_twice_clear(self, mock_check_rules):
+        self.enforcer.load_rules()
+        self.enforcer.clear()
+        # NOTE(bnemec): It's weird that we have to pass True here, but clear
+        # sets enforcer.use_conf to False, which causes load_rules to be a
+        # noop when called with no parameters.  This is probably a bug.
+        self.enforcer.load_rules(True)
+        self.assertEqual(2, mock_check_rules.call_count)
+
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_directory_twice(self, mock_check_rules):
+        self.create_config_file(
+            os.path.join('policy.d', 'a.conf'), POLICY_A_CONTENTS)
+        self.create_config_file(
+            os.path.join('policy.d', 'b.conf'), POLICY_B_CONTENTS)
+        self.enforcer.load_rules()
+        self.enforcer.load_rules()
+        self.assertEqual(1, mock_check_rules.call_count)
+        self.assertIsNotNone(self.enforcer.rules)
+
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_directory_twice_force(self, mock_check_rules):
+        self.create_config_file(
+            os.path.join('policy.d', 'a.conf'), POLICY_A_CONTENTS)
+        self.create_config_file(
+            os.path.join('policy.d', 'b.conf'), POLICY_B_CONTENTS)
+        self.enforcer.load_rules(True)
+        self.enforcer.load_rules(True)
+        self.assertEqual(2, mock_check_rules.call_count)
+        self.assertIsNotNone(self.enforcer.rules)
+
+    @mock.patch('oslo_policy.policy.Enforcer.check_rules')
+    def test_load_directory_twice_changed(self, mock_check_rules):
+        self.create_config_file(
+            os.path.join('policy.d', 'a.conf'), POLICY_A_CONTENTS)
+        self.enforcer.load_rules()
+
+        # Touch the file
+        conf_path = os.path.join(self.config_dir, os.path.join(
+            'policy.d', 'a.conf'))
+        stinfo = os.stat(conf_path)
+        os.utime(conf_path, (stinfo.st_atime + 10, stinfo.st_mtime + 10))
+
+        self.enforcer.load_rules()
+        self.assertEqual(2, mock_check_rules.call_count)
+        self.assertIsNotNone(self.enforcer.rules)
+
     def test_set_rules_type(self):
         self.assertRaises(TypeError,
                           self.enforcer.set_rules,
@@ -641,6 +707,89 @@ class EnforcerTest(base.PolicyBaseTestCase):
                           self.enforcer.authorize, 'test', {},
                           {'roles': ['test']})
 
+    def test_enforcer_accepts_context_objects(self):
+        rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
+        self.enforcer.register_default(rule)
+
+        request_context = context.RequestContext()
+        target_dict = {}
+        self.enforcer.enforce('fake_rule', target_dict, request_context)
+
+    def test_enforcer_accepts_subclassed_context_objects(self):
+        rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
+        self.enforcer.register_default(rule)
+
+        class SpecializedContext(context.RequestContext):
+            pass
+
+        request_context = SpecializedContext()
+        target_dict = {}
+        self.enforcer.enforce('fake_rule', target_dict, request_context)
+
+    def test_enforcer_rejects_non_context_objects(self):
+        rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
+        self.enforcer.register_default(rule)
+
+        class InvalidContext(object):
+            pass
+
+        request_context = InvalidContext()
+        target_dict = {}
+        self.assertRaises(
+            policy.InvalidContextObject, self.enforcer.enforce, 'fake_rule',
+            target_dict, request_context
+        )
+
+    @mock.patch.object(policy.Enforcer, '_map_context_attributes_into_creds')
+    def test_enforcer_call_map_context_attributes(self, map_mock):
+        rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
+        self.enforcer.register_default(rule)
+
+        request_context = context.RequestContext()
+        target_dict = {}
+        self.enforcer.enforce('fake_rule', target_dict, request_context)
+        map_mock.assert_called_once_with(request_context)
+
+    def test_enforcer_consolidates_context_attributes_with_creds(self):
+        request_context = context.RequestContext()
+        expected_creds = request_context.to_policy_values()
+
+        creds = self.enforcer._map_context_attributes_into_creds(
+            request_context
+        )
+
+        # We don't use self.assertDictEqual here because to_policy_values
+        # actaully returns a non-dict object that just behaves like a
+        # dictionary, but does some special handling when people access
+        # deprecated policy values.
+        for k, v in expected_creds.items():
+            self.assertEqual(expected_creds[k], creds[k])
+
+    def test_map_context_attributes_populated_system(self):
+        request_context = context.RequestContext(system_scope='all')
+        expected_creds = request_context.to_policy_values()
+        expected_creds['system'] = 'all'
+
+        creds = self.enforcer._map_context_attributes_into_creds(
+            request_context
+        )
+
+        # We don't use self.assertDictEqual here because to_policy_values
+        # actaully returns a non-dict object that just behaves like a
+        # dictionary, but does some special handling when people access
+        # deprecated policy values.
+        for k, v in expected_creds.items():
+            self.assertEqual(expected_creds[k], creds[k])
+
+    def test_enforcer_accepts_policy_values_from_context(self):
+        rule = policy.RuleDefault(name='fake_rule', check_str='role:test')
+        self.enforcer.register_default(rule)
+
+        request_context = context.RequestContext()
+        policy_values = request_context.to_policy_values()
+        target_dict = {}
+        self.enforcer.enforce('fake_rule', target_dict, policy_values)
+
 
 class EnforcerNoPolicyFileTest(base.PolicyBaseTestCase):
     def setUp(self):
@@ -692,22 +841,25 @@ class CheckFunctionTestCase(base.PolicyBaseTestCase):
 
     def test_check_explicit(self):
         rule = base.FakeCheck()
-        result = self.enforcer.enforce(rule, 'target', 'creds')
-        self.assertEqual(('target', 'creds', self.enforcer), result)
+        creds = {}
+        result = self.enforcer.enforce(rule, 'target', creds)
+        self.assertEqual(('target', creds, self.enforcer), result)
 
     def test_check_no_rules(self):
         # Clear the policy.json file created in setUp()
         self.create_config_file('policy.json', "{}")
         self.enforcer.default_rule = None
         self.enforcer.load_rules()
-        result = self.enforcer.enforce('rule', 'target', 'creds')
+        creds = {}
+        result = self.enforcer.enforce('rule', 'target', creds)
         self.assertFalse(result)
 
     def test_check_with_rule(self):
         self.enforcer.set_rules(dict(default=base.FakeCheck()))
-        result = self.enforcer.enforce('default', 'target', 'creds')
+        creds = {}
+        result = self.enforcer.enforce('default', 'target', creds)
 
-        self.assertEqual(('target', 'creds', self.enforcer), result)
+        self.assertEqual(('target', creds, self.enforcer), result)
 
     def test_check_rule_not_exist_not_empty_policy_file(self):
         # If the rule doesn't exist, then enforce() fails rather than KeyError.
@@ -716,7 +868,8 @@ class CheckFunctionTestCase(base.PolicyBaseTestCase):
         self.create_config_file('policy.json', jsonutils.dumps({"a_rule": []}))
         self.enforcer.default_rule = None
         self.enforcer.load_rules()
-        result = self.enforcer.enforce('rule', 'target', 'creds')
+        creds = {}
+        result = self.enforcer.enforce('rule', 'target', creds)
         self.assertFalse(result)
 
     def test_check_raise_default(self):
@@ -724,15 +877,20 @@ class CheckFunctionTestCase(base.PolicyBaseTestCase):
         # raised.
         self.enforcer.set_rules(dict(default=_checks.FalseCheck()))
 
-        self.assertRaises(policy.PolicyNotAuthorized, self.enforcer.enforce,
-                          'rule', 'target', 'creds', True)
+        creds = {}
+        self.assertRaisesRegex(policy.PolicyNotAuthorized,
+                               " is disallowed by policy",
+                               self.enforcer.enforce,
+                               'rule', 'target', creds, True)
 
     def test_check_raise_custom_exception(self):
         self.enforcer.set_rules(dict(default=_checks.FalseCheck()))
 
+        creds = {}
         exc = self.assertRaises(
-            MyException, self.enforcer.enforce, 'rule', 'target', 'creds',
-            True, MyException, 'arg1', 'arg2', kw1='kwarg1', kw2='kwarg2')
+            MyException, self.enforcer.enforce, 'rule', 'target', creds,
+            True, MyException, 'arg1', 'arg2', kw1='kwarg1',
+            kw2='kwarg2')
         self.assertEqual(('arg1', 'arg2'), exc.args)
         self.assertEqual(dict(kw1='kwarg1', kw2='kwarg2'), exc.kwargs)
 
@@ -772,7 +930,7 @@ class BaseCheckTypesTestCase(base.PolicyBaseTestCase):
 class RuleDefaultTestCase(base.PolicyBaseTestCase):
     def test_rule_is_parsed(self):
         opt = policy.RuleDefault(name='foo', check_str='rule:foo')
-        self.assertTrue(isinstance(opt.check, _checks.BaseCheck))
+        self.assertIsInstance(opt.check, _checks.BaseCheck)
         self.assertEqual('rule:foo', str(opt.check))
 
     def test_str(self):
@@ -832,3 +990,356 @@ class RuleDefaultTestCase(base.PolicyBaseTestCase):
         opt1 = policy.RuleDefault(name='foo', check_str='rule:foo')
         opt2 = RuleDefaultSub(name='bar', check_str='rule:foo')
         self.assertNotEqual(opt1, opt2)
+
+    def test_create_opt_with_scope_types(self):
+        scope_types = ['project']
+        opt = policy.RuleDefault(
+            name='foo',
+            check_str='role:bar',
+            scope_types=scope_types
+        )
+        self.assertEqual(opt.scope_types, scope_types)
+
+    def test_create_opt_with_scope_type_strings_fails(self):
+        self.assertRaises(
+            ValueError,
+            policy.RuleDefault,
+            name='foo',
+            check_str='role:bar',
+            scope_types='project'
+        )
+
+    def test_create_opt_with_multiple_scope_types(self):
+        opt = policy.RuleDefault(
+            name='foo',
+            check_str='role:bar',
+            scope_types=['project', 'system']
+        )
+
+        self.assertEqual(opt.scope_types, ['project', 'system'])
+
+    def test_ensure_scope_types_are_unique(self):
+        self.assertRaises(
+            ValueError,
+            policy.RuleDefault,
+            name='foo',
+            check_str='role:bar',
+            scope_types=['project', 'project']
+        )
+
+
+class DocumentedRuleDefaultDeprecationTestCase(base.PolicyBaseTestCase):
+
+    def test_deprecate_a_policy_check_string(self):
+        deprecated_rule = policy.DeprecatedRule(
+            name='foo:create_bar',
+            check_str='role:fizz'
+        )
+
+        rule_list = [policy.DocumentedRuleDefault(
+            name='foo:create_bar',
+            check_str='role:bang',
+            description='Create a bar.',
+            operations=[{'path': '/v1/bars', 'method': 'POST'}],
+            deprecated_rule=deprecated_rule,
+            deprecated_reason='"role:bang" is a better default',
+            deprecated_since='N'
+        )]
+        enforcer = policy.Enforcer(self.conf)
+        enforcer.register_defaults(rule_list)
+        expected_msg = (
+            'Policy "foo:create_bar":"role:fizz" was deprecated in N in favor '
+            'of "foo:create_bar":"role:bang". Reason: "role:bang" is a better '
+            'default. Either ensure your deployment is ready for the new '
+            'default or copy/paste the deprecated policy into your policy '
+            'file and maintain it manually.'
+        )
+
+        with mock.patch('warnings.warn') as mock_warn:
+            enforcer.load_rules()
+            mock_warn.assert_called_once_with(expected_msg)
+
+    def test_deprecate_a_policy_name(self):
+        deprecated_rule = policy.DeprecatedRule(
+            name='foo:bar',
+            check_str='role:baz'
+        )
+
+        rule_list = [policy.DocumentedRuleDefault(
+            name='foo:create_bar',
+            check_str='role:baz',
+            description='Create a bar.',
+            operations=[{'path': '/v1/bars/', 'method': 'POST'}],
+            deprecated_rule=deprecated_rule,
+            deprecated_reason=(
+                '"foo:bar" is not granular enough. If your deployment has '
+                'overridden "foo:bar", ensure you override the new policies '
+                'with same role or rule. Not doing this will require the '
+                'service to assume the new defaults for "foo:bar:create", '
+                '"foo:bar:update", "foo:bar:list", and "foo:bar:delete", '
+                'which might be backwards incompatible for your deployment'
+            ),
+            deprecated_since='N'
+        )]
+        expected_msg = (
+            'Policy "foo:bar":"role:baz" was deprecated in N in favor of '
+            '"foo:create_bar":"role:baz". Reason: "foo:bar" is not granular '
+            'enough. If your deployment has overridden "foo:bar", ensure you '
+            'override the new policies with same role or rule. Not doing this '
+            'will require the service to assume the new defaults for '
+            '"foo:bar:create", "foo:bar:update", "foo:bar:list", and '
+            '"foo:bar:delete", which might be backwards incompatible for your '
+            'deployment. Either ensure your deployment is ready for the new '
+            'default or copy/paste the deprecated policy into your policy '
+            'file and maintain it manually.'
+        )
+
+        rules = jsonutils.dumps({'foo:bar': 'role:bang'})
+        self.create_config_file('policy.json', rules)
+        enforcer = policy.Enforcer(self.conf)
+        enforcer.register_defaults(rule_list)
+
+        with mock.patch('warnings.warn') as mock_warn:
+            enforcer.load_rules(True)
+            mock_warn.assert_called_once_with(expected_msg)
+
+    def test_deprecate_a_policy_for_removal_logs_warning_when_overridden(self):
+        rule_list = [policy.DocumentedRuleDefault(
+            name='foo:bar',
+            check_str='role:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_for_removal=True,
+            deprecated_reason=(
+                '"foo:bar" is no longer a policy used by the service'
+            ),
+            deprecated_since='N'
+        )]
+        expected_msg = (
+            'Policy "foo:bar":"role:baz" was deprecated for removal in N. '
+            'Reason: "foo:bar" is no longer a policy used by the service. Its '
+            'value may be silently ignored in the future.'
+        )
+        rules = jsonutils.dumps({'foo:bar': 'role:bang'})
+        self.create_config_file('policy.json', rules)
+        enforcer = policy.Enforcer(self.conf)
+        enforcer.register_defaults(rule_list)
+
+        with mock.patch('warnings.warn') as mock_warn:
+            enforcer.load_rules()
+            mock_warn.assert_called_once_with(expected_msg)
+
+    def test_deprecate_a_policy_for_removal_does_not_log_warning(self):
+        # We should only log a warning for operators if they are supplying an
+        # override for a policy that is deprecated for removal.
+        rule_list = [policy.DocumentedRuleDefault(
+            name='foo:bar',
+            check_str='role:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_for_removal=True,
+            deprecated_reason=(
+                '"foo:bar" is no longer a policy used by the service'
+            ),
+            deprecated_since='N'
+        )]
+        enforcer = policy.Enforcer(self.conf)
+        enforcer.register_defaults(rule_list)
+
+        with mock.patch('warnings.warn') as mock_warn:
+            enforcer.load_rules()
+            mock_warn.assert_not_called()
+
+    def test_deprecated_policy_for_removal_must_include_deprecated_since(self):
+        self.assertRaises(
+            ValueError,
+            policy.DocumentedRuleDefault,
+            name='foo:bar',
+            check_str='rule:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_for_removal=True,
+            deprecated_reason='Some reason.'
+        )
+
+    def test_deprecated_policy_must_include_deprecated_since(self):
+        deprecated_rule = policy.DeprecatedRule(
+            name='foo:bar',
+            check_str='rule:baz'
+        )
+
+        self.assertRaises(
+            ValueError,
+            policy.DocumentedRuleDefault,
+            name='foo:bar',
+            check_str='rule:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_rule=deprecated_rule,
+            deprecated_reason='Some reason.'
+        )
+
+    def test_deprecated_rule_requires_deprecated_rule_object(self):
+        self.assertRaises(
+            ValueError,
+            policy.DocumentedRuleDefault,
+            name='foo:bar',
+            check_str='rule:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_rule='foo:bar',
+            deprecated_reason='Some reason.'
+        )
+
+    def test_deprecated_policy_must_include_deprecated_reason(self):
+        self.assertRaises(
+            ValueError,
+            policy.DocumentedRuleDefault,
+            name='foo:bar',
+            check_str='rule:baz',
+            description='Create a foo.',
+            operations=[{'path': '/v1/foos/', 'method': 'POST'}],
+            deprecated_for_removal=True,
+            deprecated_since='N'
+        )
+
+
+class DocumentedRuleDefaultTestCase(base.PolicyBaseTestCase):
+
+    def test_contain_operations(self):
+        opt = policy.DocumentedRuleDefault(
+            name='foo', check_str='rule:foo', description='foo_api',
+            operations=[{'path': '/foo/', 'method': 'GET'}])
+
+        self.assertEqual(1, len(opt.operations))
+
+    def test_multiple_operations(self):
+        opt = policy.DocumentedRuleDefault(
+            name='foo', check_str='rule:foo', description='foo_api',
+            operations=[{'path': '/foo/', 'method': 'GET'},
+                        {'path': '/foo/', 'method': 'POST'}])
+
+        self.assertEqual(2, len(opt.operations))
+
+    def test_description_not_empty(self):
+        invalid_desc = ''
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description=invalid_desc,
+                          operations=[{'path': '/foo/', 'method': 'GET'}])
+
+    def test_operation_not_empty_list(self):
+        invalid_op = []
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+    def test_operation_must_be_list(self):
+        invalid_op = 'invalid_op'
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+    def test_operation_must_be_list_of_dicts(self):
+        invalid_op = ['invalid_op']
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+    def test_operation_must_have_path(self):
+        invalid_op = [{'method': 'POST'}]
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+    def test_operation_must_have_method(self):
+        invalid_op = [{'path': '/foo/path/'}]
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+    def test_operation_must_contain_method_and_path_only(self):
+        invalid_op = [{'path': '/some/path/',
+                       'method': 'GET',
+                       'break': 'me'}]
+        self.assertRaises(policy.InvalidRuleDefault,
+                          policy.DocumentedRuleDefault,
+                          name='foo',
+                          check_str='rule:foo',
+                          description='foo_api',
+                          operations=invalid_op)
+
+
+class EnforcerCheckRulesTest(base.PolicyBaseTestCase):
+    def setUp(self):
+        super(EnforcerCheckRulesTest, self).setUp()
+
+    def test_no_violations(self):
+        self.create_config_file('policy.json', POLICY_JSON_CONTENTS)
+        self.enforcer.load_rules(True)
+        self.assertTrue(self.enforcer.check_rules(raise_on_violation=True))
+
+    def test_undefined_rule(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertFalse(self.enforcer.check_rules())
+
+    def test_undefined_rule_raises(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertRaises(policy.InvalidDefinitionError,
+                          self.enforcer.check_rules, raise_on_violation=True)
+
+    def test_cyclical_rules(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar', 'bar': 'rule:foo'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertFalse(self.enforcer.check_rules())
+
+    def test_cyclical_rules_raises(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar', 'bar': 'rule:foo'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertRaises(policy.InvalidDefinitionError,
+                          self.enforcer.check_rules, raise_on_violation=True)
+
+    def test_complex_cyclical_rules_false(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar',
+                                 'bar': 'rule:baz and role:admin',
+                                 'baz': 'rule:foo or role:user'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertFalse(self.enforcer.check_rules())
+
+    def test_complex_cyclical_rules_true(self):
+        rules = jsonutils.dumps({'foo': 'rule:bar or rule:baz',
+                                 'bar': 'role:admin',
+                                 'baz': 'rule:bar or role:user'})
+        self.create_config_file('policy.json', rules)
+        self.enforcer.load_rules(True)
+
+        self.assertTrue(self.enforcer.check_rules())
